@@ -3,7 +3,7 @@
 #
 # This file is part of scqubits.
 #
-#    Copyright (c) 2019, Jens Koch and Peter Groszkowski
+#    Copyright (c) 2019 and later, Jens Koch and Peter Groszkowski
 #    All rights reserved.
 #
 #    This source code is licensed under the BSD-style license found in the
@@ -21,8 +21,8 @@ class TestFluxQubit(StandardTests):
     def setup_class(cls):
         cls.qbt = None
         cls.qbt_type = FluxQubit
-        cls.file_str = 'fluxqubit'
-        cls.op1_str = 'n_1_operator'
-        cls.op2_str = 'n_2_operator'
-        cls.param_name = 'flux'
+        cls.file_str = "fluxqubit"
+        cls.op1_str = "n_1_operator"
+        cls.op2_str = "n_2_operator"
+        cls.param_name = "flux"
         cls.param_list = np.linspace(0.45, 0.55, 50)
